(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: translate.ml,v 1.6 1998/08/16 19:22:55 xleroy Exp $ *)

(* Translation from 1.07 syntax to 2.00 syntax *)

open Printf
open Location
open Longident
open Asttypes
open Parsetree

let ic = ref stdin
let oc = ref stdout

let oprintf fmt = fprintf !oc fmt

(* Copy a chunk of text from ic to oc *)

let copy_buffer = String.create 1024

let copy_chunk loc =
  assert (loc.loc_start <= loc.loc_end);
  seek_in !ic loc.loc_start;
  let toread = ref (loc.loc_end - loc.loc_start) in
  while !toread > 0 do
    let read =
      input !ic copy_buffer 0 (min !toread (String.length copy_buffer)) in
    output !oc copy_buffer 0 read;
    toread := !toread - read
  done

(* Copy from last point to beginning of a location *)

let last_copy_point = ref 0

let copy_upto loc =
  assert (loc.loc_start >= !last_copy_point);
  copy_chunk {loc_start = !last_copy_point; loc_end = loc.loc_start};
  last_copy_point := loc.loc_start

let start_next_copy loc =
  last_copy_point := loc.loc_end

let copy_to_end () =
  copy_chunk {loc_start = !last_copy_point; loc_end = in_channel_length !ic}

(* Helpers *)

let rec output_longident oc = function
    Lident s ->
      output_string oc s
  | Ldot(l, s) ->
      output_longident oc l; output_char oc '.'; output_string oc s
  | Lapply(l1, l2) ->
      output_longident oc l1; output_char oc '(';
      output_longident oc l2; output_char oc ')'

(* Translate class signatures *)

let transl_field_decl = function
    Pctf_val(id, priv, mut, ty, loc) ->
      copy_upto loc;
      if priv = Private then
        Location.print_warning loc "no more 'private' annotations on instance variables, use a class type constraint instead";
      oprintf "val %s%s%s"
              (if mut = Mutable then "mutable " else "")
              (if priv = Private then "(*private*) " else "")
              id;
      begin match ty with
        None ->
          Location.print_warning loc "no more instance variables without type; add correct type there or remove variable redeclaration entirely"
      | Some t -> oprintf " : "; copy_chunk t.ptyp_loc
      end;
      start_next_copy loc
  | Pctf_virt(id, priv, ty, loc) ->
      copy_upto loc;
      oprintf "method %svirtual %s : "
              (if priv = Private then "private " else "") id;
      copy_chunk ty.ptyp_loc;
      start_next_copy loc
  | Pctf_meth(id, priv, ty, loc) ->
      copy_upto loc;
      oprintf "method %s%s : "
              (if priv = Private then "private " else "") id;
      copy_chunk ty.ptyp_loc;
      start_next_copy loc
  | Pctf_inher(id, ty_args, loc) ->
      copy_upto loc;
      oprintf "inherit";
      begin match ty_args with
        [] -> ()
      | t1 :: tl ->
          oprintf " ["; copy_chunk t1.ptyp_loc;
          List.iter (fun t -> oprintf ", "; copy_chunk t.ptyp_loc) tl;
          oprintf "]"
      end;
      oprintf " %a" output_longident id;
      start_next_copy loc

let transl_classtype cty lastone =
  copy_upto cty.pcty_loc;
  if cty.pcty_kind = Virtual then oprintf " virtual";
  if cty.pcty_closed = Closed then begin
    Location.print_warning cty.pcty_loc "'closed' classes are no longer supported";
    oprintf " (*closed*)"
  end;
  begin match cty.pcty_param with
    ([], _) -> ()
  | (p1 :: pl, _) ->
      oprintf " ['%s" p1;
      List.iter (fun p -> oprintf ", '%s" p) pl;
      oprintf "]"
  end;
  oprintf " %s :" cty.pcty_name;
  List.iter
    (fun ty -> oprintf " ("; copy_chunk ty.ptyp_loc; oprintf ") ->")
    cty.pcty_args;
  oprintf " object";
  begin match cty.pcty_self with
    None -> ()
  | Some self -> oprintf " ('%s)" self
  end;
  start_next_copy cty.pcty_equal_loc;
  (* Constraints can be copied as is *)
  List.iter transl_field_decl cty.pcty_field;
  if not lastone then oprintf "\n  end"
  (* start_next_copy cty.pcty_loc *)

let rec transl_classtypes = function
    [] -> assert false
  | [cls] -> transl_classtype cls true
  | cls1 :: clsl -> transl_classtype cls1 false; transl_classtypes clsl

(* Translate signature and module types *)

let rec transl_module_type m =
  match m.pmty_desc with
    Pmty_signature sg -> List.iter transl_signature_item sg
  | Pmty_functor(s, mty1, mty2) ->
      transl_module_type mty1; transl_module_type mty2
  | Pmty_with(mty, cstr) -> transl_module_type mty
  | _ -> ()

and transl_signature_item i =
  match i.psig_desc with
    Psig_class clstys -> transl_classtypes clstys
  | Psig_module(s, mty) -> transl_module_type mty
  | Psig_modtype(s, Pmodtype_manifest mty) -> transl_module_type mty
  | _ -> ()

(* Translate a field declaration *)

let transl_field = function
    Pcf_val(id, priv, mut, init, loc) ->
      copy_upto loc;
      if priv = Private then
        Location.print_warning loc "no more 'private' annotations on instance variables, use a class type constraint instead";
      begin match init with
        Some { pexp_desc = Pexp_ident(Lident s) }
        when mut = Immutable && s = id ->
          Location.print_warning loc
           (sprintf "`val %s = %s' no longer needed, commenting it out" id id);
         oprintf "(* val %s = %s *)" id id
      | _ ->
        oprintf "val %s%s%s"
                (if mut = Mutable then "mutable " else "")
                (if priv = Private then "(*private*) " else "")
                id;
        begin match init with
          None ->
            Location.print_warning loc "no more instance variables without initialization; add correct initialization there or remove variable redeclaration entirely"
        | Some e -> oprintf " = "; copy_chunk e.pexp_loc
        end
      end;      
      start_next_copy loc
  | Pcf_virt(id, priv, ty, loc) ->
      copy_upto loc;
      oprintf "method %svirtual %s : "
              (if priv = Private then "private " else "") id;
      copy_chunk ty.ptyp_loc;
      start_next_copy loc
  | Pcf_meth(id, priv, body, loc, body_loc) ->
      copy_upto loc;
      oprintf "method %s%s "
              (if priv = Private then "private " else "") id;
      copy_chunk body_loc;
      start_next_copy loc
  | Pcf_inher(id, ty_args, args, super, loc) ->
      copy_upto loc;
      oprintf "inherit";
      begin match ty_args with
        [] -> ()
      | t1 :: tl ->
          oprintf " ["; copy_chunk t1.ptyp_loc;
          List.iter (fun t -> oprintf ", "; copy_chunk t.ptyp_loc) tl;
          oprintf "]"
      end;
      oprintf " %a" output_longident id;
      List.iter (fun e -> oprintf " "; copy_chunk e.pexp_loc) args;
      begin match super with
        None -> ()
      | Some s -> oprintf " as %s" s
      end;
      start_next_copy loc

(* Translate class definitions *)

let transl_class cls lastone =
  copy_upto cls.pcl_loc;
  if cls.pcl_kind = Virtual then oprintf " virtual";
  if cls.pcl_closed = Closed then oprintf " (*closed*)";
  begin match cls.pcl_param with
    ([], _) -> ()
  | (p1 :: pl, _) ->
      oprintf " ['%s" p1;
      List.iter (fun p -> oprintf ", '%s" p) pl;
      oprintf "]"
  end;
  oprintf " %s" cls.pcl_name;
  List.iter (fun pat -> oprintf " "; copy_chunk pat.ppat_loc) cls.pcl_args;
  oprintf " =\n object";
  begin match (cls.pcl_self, cls.pcl_self_ty) with
    (None, None) -> ()
  | (Some self, None) -> oprintf " (%s)" self
  | (None, Some selfty) -> oprintf " (_self : '%s)" selfty
  | (Some self, Some selfty) -> oprintf " (%s : '%s)" self selfty
  end;
  start_next_copy cls.pcl_equal_loc;
  (* Constraints can be copied as is *)
  List.iter transl_field cls.pcl_field;
  if not lastone then oprintf "\n  end"
(*  start_next_copy cls.pcl_loc *)

let rec transl_classes = function
    [] -> assert false
  | [cls] -> transl_class cls true
  | cls1 :: clsl -> transl_class cls1 false; transl_classes clsl

(* Translate structures and module expressions *)

let rec transl_module_expr m =
  match m.pmod_desc with
    Pmod_structure str -> List.iter transl_structure_item str
  | Pmod_functor(str, mty, m) -> transl_module_type mty; transl_module_expr m
  | Pmod_apply(m1, m2) -> transl_module_expr m1; transl_module_expr m2
  | Pmod_constraint(m, mty) ->
      if m.pmod_loc.loc_end <= mty.pmty_loc.loc_start
      then (transl_module_expr m; transl_module_type mty)
      else (transl_module_type mty; transl_module_expr m)
  | _ -> ()

and transl_structure_item i =
  match i.pstr_desc with
    Pstr_class clsdefs -> transl_classes clsdefs
  | Pstr_module(s, m) -> transl_module_expr m
  | Pstr_modtype(s, mty) -> transl_module_type mty
  | _ -> ()

(* Translate an implementation file *)

let transl_implementation i o impl =
  ic := i; oc := o; last_copy_point := 0;
  List.iter transl_structure_item impl;
  copy_to_end()

(* Translate an interface file *)

let transl_interface i o intf =
  ic := i; oc := o; last_copy_point := 0;
  List.iter transl_signature_item intf;
  copy_to_end()

